"""
.. module:: input

:Synopsis: Input-related functions
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os
from collections import OrderedDict as odict
import numpy as np
import datetime
from numbers import Number
from getdist import MCSamples
from copy import deepcopy
from importlib import import_module

# Local
from cobaya.conventions import package, subfolders, _defaults_file, _params, _p_label
from cobaya.conventions import _prior, _theory, _likelihood, _sampler, _external
from cobaya.tools import get_folder
from cobaya.yaml_custom import yaml_load_file
from cobaya.log import HandledException
from cobaya.parametrisation import is_sampled_param, is_derived_param

# Logger
import logging
log = logging.getLogger(__name__)


def load_input(input_file):
    """
    Loads general info, and splits it into the right parts.
    """
    file_name, extension = os.path.splitext(input_file)
    file_name = os.path.basename(file_name)
    if extension in (".yaml",".yml"):
        info = yaml_load_file(input_file)
        # if output_prefix not defined, default to input_file name (sans ext.) as prefix;
        if "output_prefix" not in info:
            info["output_prefix"] = file_name
        # warn if no output, since we are in shell-invocation mode.
        elif info["output_prefix"] == None:
            log.warning("WARNING: Output explicitly supressed with 'ouput_prefix: null'")
    else:
        log.error("Extension '%s' of input file '%s' not recognised.", extension, input_file)
        raise HandledException
    return info

# MPI wrapper for loading the input info
def load_input_MPI(input_file):
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    # Load input (only one process does read the input file)
    if rank == 0: 
        info = load_input(input_file)
    else:
        info = None
    info = comm.bcast(info, root=0)
    return info

def get_modules(*infos):
    """Returns modules all requested as an odict ``{kind: set([modules])}``.
    Priors are not included."""
    modules = odict()
    for info in infos:
        for field in [_theory, _likelihood, _sampler]:
            if not field in modules:
                modules[field] = set()
            modules[field] |= (lambda v: set(v) if v else set())(info.get(field))
            modules[field] = modules[field].difference(set([None]))
    # pop empty fields
    for k,v in modules.iteritems():
        if not v:
            modules.pop(k)
    return modules

def get_full_info(info):
    """
    Creates an updated info starting from the defaults for each module and updating it
    with the input info.
    """
    # Don't modify the original input!
    input_info = deepcopy(info)
    # Creates an equivalent info using only the defaults
    full_info = odict()
    default_params_info = odict()
    default_prior_info = odict()
    modules = get_modules(input_info)
    for block in modules:
        full_info[block] = odict()
        for module in modules[block]:
            # Start with the default class options
            full_info[block][module] = deepcopy(getattr(
                import_module(package+"."+block, package=package), "class_options", {}))
            # Go on with defaults
            path_to_defaults = os.path.join(get_folder(module, block), _defaults_file)
            try:
                default_module_info = yaml_load_file(path_to_defaults)
            except IOError:
                # probably an external module
                default_module_info = {block: {module: {}}}
                log.debug("Module %s:%s does not have a defaults file. "%(block, module)+
                          "Maybe it is an external module.")
            try:
                full_info[block][module].update(default_module_info[block][module] or {})
            except KeyError:
                log.error("The defaults file for '%s' should be structured as %s:%s:{[options]}."
                          %(module, block, module))
                raise HandledException
            # Update the options with the input file
            # Consistency is checked only up to first level! (i.e. subkeys may not match)
            # First deal with cases "no options" and "external function"
            try:
                input_info[block][module] = input_info[block][module] or {}
            except TypeError:
                log.error("Your input info is not well formatted at the '%s' block. "
                          "It must be a dictionary {'%s':{options}, ...}. ", block, block)
                raise HandledException
            if not hasattr(input_info[block][module], "get"):
                input_info[block][module] = {_external: input_info[block][module]}
            options_not_recognised = (set(input_info[block][module])
                                      .difference(set([_external]))
                                      .difference(set(full_info[block][module])))
            if options_not_recognised:
                log.error("'%s' does not recognise some options: '%r'. "
                          "To see the allowed options, check out the file '%s'",
                          module, tuple(options_not_recognised), path_to_defaults)
                raise HandledException
            full_info[block][module].update(input_info[block][module])
            # Store default parameters and priors of class, and save to combine later
            if block == _likelihood:
                params_info = default_module_info.get(_params, {})
                full_info[block][module].update({_params:params_info})
                default_params_info[module] = params_info
                default_prior_info[module] = default_module_info.get(_prior, {})
    # Add priors info, after the necessary checks
    if _prior in input_info or any(default_prior_info.values()):
        full_info[_prior] = input_info.get(_prior, odict())
    for prior_info in default_prior_info.values():
        for name, prior in prior_info.iteritems():
            if full_info[_prior].get(name, prior) != prior:
                log.error("Two different priors have been defined with the same name: '%s'.",name)
                raise HandledException
            full_info[_prior][name] = prior
    # Add parameters info, after the necessary updates and checks
    full_info[_params] = merge_params_info(input_info.get(_params, {}), defaults=default_params_info)
    # Rest of the options
    for k,v in input_info.iteritems():
        if not k in full_info:
            full_info[k] = v
    return full_info

def merge_params_info(params_info, defaults=None):
    """
    Merges input and default parameters info, after performing some consistency checks.
    """
    # First, merge defaults. Impose multiple defined (=shared) parameters have equal info
    defaults_merged = odict()
    for lik, params in defaults.iteritems():
        for p, info in params.iteritems():
            # if already there, check consistency
            if p in defaults_merged:
                log.debug("Parameter '%s' multiply defined.", p)
                if info != defaults_merged[p]:
                    log.error("Parameter '%s' multiply defined, but inconsistent info: "
                              "For likelihood '%s' is '%r', but for some other likelihood "
                              "it was '%r'. Check your defaults!",
                              p, lik, info, defaults_merged[p])
                    raise HandledException
            defaults_merged[p] = info
    # Combine with the input parameter info (make sure the theory parameters come first)
    params_info_copy = deepcopy(params_info)
    theory_params = params_info_copy.pop(_theory, None)
    info_updated = odict()
    if theory_params:
        info_updated.update({_theory: theory_params})
    info_updated.update(defaults_merged)
    info_updated.update(params_info)
    # Inherit labels (for sampled and derived) and min/max (just for derived params)
    getter = lambda info, key: getattr(info, "get", lambda x: None)(key)
    for p in defaults_merged:
        default_label = getter(defaults_merged[p], _p_label)
        if (default_label and
            (is_sampled_param(info_updated[p]) or is_derived_param(info_updated[p]))):
            info_updated[p][_p_label] = info_updated[p].get(_p_label) or default_label
        limits = ["min", "max"]
        default_limits = odict([[lim,getter(defaults_merged[p], lim)] for lim in limits])
        if default_limits.values() != [None, None] and is_derived_param(info_updated[p]):
            info_updated[p].update(default_limits)
    return info_updated



"""
.. module:: samplers.mcmc.proposal

:Synopsis: proposal distributions
:Author: Antony Lewis (from CosmoMC)

Using the covariance matrix to give the proposal directions typically
significantly increases the acceptance rate and gives faster movement
around parameter space.

We generate a random basis in the eigenvectors, then cycle through them
proposing changes to each, then generate a new random basis.
The distance proposal in the random direction is given by a two-D Gaussian
radial function mixed with an exponential, which is quite robust to wrong width estimates

See http://cosmologist.info/notes/CosmoMC.pdf
"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import numpy as np
import scipy
import numpy

# Local
from cobaya.log import HandledException

# Logger
import logging
log = logging.getLogger(__name__)


class IndexCycler(object):
    def __init__(self, n):
        self.n = n
        self.loopix = -1


class CyclicIndexRandomizer(IndexCycler):
    def next(self):
        """
        Get the next random index

        :return: index
        """
        self.loopix = (self.loopix + 1) % self.n
        if self.loopix == 0:
            self.indices = np.random.permutation(self.n)
        return self.indices[self.loopix]


class RandDirectionProposer(IndexCycler):
    def propose_vec(self, scale):
        """
        propose a random n-dimension vector

        :param scale: units for the distance
        :return: array with vector
        """
        self.loopix = (self.loopix + 1) % self.n
        if self.loopix == 0:
            if self.n > 1:
                self.R = scipy.stats.special_ortho_group.rvs(self.n)
            else:
                self.R = np.eye(1)*numpy.random.choice((-1,1))
        return self.R[:, self.loopix] * self.propose_r() * scale

    def propose_r(self):
        """
        Radial proposal. By default a mixture of an exponential and 2D Gaussian radial proposal
        (to make wider tails and more mass near zero, so more robust to scale misestimation)
        :return: random distance (unit scale)
        """
        if np.random.uniform() < 0.33:
            return np.random.exponential()
        else:
            return np.linalg.norm(np.random.normal(size=min(self.n, 2)))


class BlockProposer(RandDirectionProposer):
    def update_params(self, params, vec):
        params[self.params_changed] += self.mapping_matrix.dot(vec)


class CovmatProposer(object):
    def set_covariance(self):
        pass


class BlockedProposer(CovmatProposer):
    def __init__(self, parameter_blocks, params_used, slow_block_max,
                 oversample_fast=1, propose_scale=2.4):

        """
        Proposal density for fast and slow parameters, where parameters are
        grouped into blocks which are changed at the same time.

        :param parameter_blocks: list of arrays of parameter indices in each block
        :param params_used: array of indices of actual parameters that are varied
        :param slow_block_max: index of block which is the last which is slow
        :param oversample_fast: factor by which to oversample fast parameters
        :param propose_scale: overal scale for the proposal
        """
        self.oversample_fast = oversample_fast
        self.propose_scale = propose_scale
        self.fast_ix = 0
        # total number and number of slow parameters
        n_all = 0
        n_slow = 0
        used_blocks_indices = []
        for i, block in enumerate(parameter_blocks):
            n_block = len(block)
            if n_block:
                used_blocks_indices.append(i)
                n_all += n_block
                if i <= slow_block_max:
                    n_slow += n_block
        self.all  = CyclicIndexRandomizer(n_all)
        self.slow = CyclicIndexRandomizer(n_slow)
        self.fast = CyclicIndexRandomizer(n_all - n_slow)
        # Creating the proposers
        self.proposer = []
        self.proposer_for_index = range(n_all)
        self.used_param_indices = np.zeros(n_all, dtype=int)
        # parameter index
        ix = 0
        for i_block in used_blocks_indices:
            pars = parameter_blocks[i_block]
            bp = BlockProposer(len(pars))
            bp.block_start = ix
            bp.used_param_indices = pars[:] # this is a copy
            self.proposer.append(bp)
            self.proposer_for_index[ix:ix + bp.n] = [bp for _ in range(bp.n)]
            self.used_param_indices[ix:ix + bp.n] = pars
            ix += bp.n
        for i_block, bp in enumerate(self.proposer):
            bp.used_params_changed = self.used_param_indices[bp.block_start:]
            # ensuring params_used is an array, to be able to select with list of indices
            bp.params_changed = np.array(params_used)[bp.used_params_changed]

    def set_covariance(self, propose_matrix):
        """
        Take covariance of used parameters (propose_matrix), and construct orthonormal parameters
        where orthonormal parameters are grouped in blocks by speed, so changes in slowest block
        changes slow and fast parameters, but changes in the fastest block only changes fast parameters

        :param propose_matrix: covariance matrix for used parameters
        """
        if not (np.allclose(propose_matrix.T,propose_matrix) and
                np.all(np.linalg.eigvals(propose_matrix) > 0)):
            log.error("The given covmat is not a positive-definite, "
                      "symmetric square matrix.")
            raise HandledException
        self.propose_matrix = propose_matrix.copy()
        sigmas = np.sqrt(np.diag(propose_matrix))
        corr = propose_matrix[np.ix_(self.used_param_indices, self.used_param_indices)]
        # before: "in range(self.size)", UNDEFINED!!!
        for i in range(len(self.used_param_indices)):
            s = np.sqrt(corr[i, i])
            corr[i, :] /= s
            corr[:, i] /= s
        L = np.linalg.cholesky(corr)
        for i, bp in enumerate(self.proposer):
            bp.mapping_matrix = np.empty((len(bp.used_params_changed), bp.n))
            for j, par in enumerate(bp.used_params_changed):
                bp.mapping_matrix[j, :] = (sigmas[par] *
                                           L[bp.block_start + j, bp.block_start:bp.block_start+bp.n])

    def get_covariance(self):
        return self.propose_matrix.copy()

    def get_block_proposal(self, P, bp):
        bp.update_params(P, bp.propose_vec(self.propose_scale))

    def get_proposal(self, P):
        if self.fast_ix:
            self.get_proposal_fast(P)
            self.fast_ix -= 1
        else:
            if self.all.next() > (self.slow.n - 1):
                self.get_proposal_fast(P)
                self.fast_ix = self.oversample_fast - 1
            else:
                self.get_proposal_slow(P)

    def get_proposal_slow(self, P):
        self.get_block_proposal(P, self.proposer_for_index[self.slow.next()])

    def get_proposal_fast(self, P):
        self.get_block_proposal(P, self.proposer_for_index[self.slow.n + self.fast.next()])

    def get_proposal_fast_delta(self,P):
        P[:]=0
        self.get_proposal_fast(P)
